#include <Arduino.h>
#include <ESP8266WiFi.h>
#include <WiFiClient.h>
#include <DHTesp.h>
#include <NTPClient.h>
#include <WiFiUdp.h>
#include <ArduinoOTA.h>
#include "Adafruit_MQTT.h"
#include "Adafruit_MQTT_Client.h"

// --------- Information to Connect to Adafruit IO ------------------------------------------------------
#define AIO_SERVER      "io.adafruit.com"
#define AIO_SERVERPORT  1883                   // use 8883 for SSL
#define AIO_USERNAME    "iu2frl"
#define AIO_KEY         ""
#define TEMP_FEED       "/feeds/stazione-meteo.temperatura"
#define HUM_FEED        "/feeds/stazione-meteo.umidita"
#define WIND_FEED       "/feeds/stazione-meteo.vento"
#define RAF_FEED       "/feeds/stazione-meteo.raffica"

// --------- Frequenza aggiornamenti --------------------------------------------------------------------
#define WIND_INTERVAL 30*1000 // Secondi
#define IOTTEMP_INTERVAL 5*60*1000 // Minuti
#define IOTWIND_INTERVAL 1*60*1000 // Minuti

// --------- PINOUT -------------------------------------------
#define pinTemp D1 // Pin DHT-11
#define pinVento D2 // Pin Anemometro

// ----------- Parametri WiFi --------------------------------------------------------------------------
#define MySSID "" // Change to your Wifi network SSID
#define MyWifiPassword "" // Change to your Wifi network password

#define ragAn 0.08 // Raggio anemometro in m
#define filtro 0.95 // Filtro passa basso

// --------- Stringhe IOT -------------------------------------------------------------------------------
//WebSocketsClient webSocket;
WiFiClient client;
Adafruit_MQTT_Client mqtt(&client, AIO_SERVER, AIO_SERVERPORT, AIO_USERNAME, AIO_KEY);
Adafruit_MQTT_Publish FeedTemperatura = Adafruit_MQTT_Publish(&mqtt, AIO_USERNAME TEMP_FEED);
Adafruit_MQTT_Publish FeedUmidita = Adafruit_MQTT_Publish(&mqtt, AIO_USERNAME HUM_FEED);
Adafruit_MQTT_Publish FeedVento = Adafruit_MQTT_Publish(&mqtt, AIO_USERNAME WIND_FEED);
Adafruit_MQTT_Publish FeedRaffica = Adafruit_MQTT_Publish(&mqtt, AIO_USERNAME RAF_FEED);

//-------------------------------------------------------------------------------------------------------
// --------- Non toccare da qui in poi ------------------------------------------------------------------
//-------------------------------------------------------------------------------------------------------
float humidity;
float temperature;
float temperaturefh;
double wind;
double windmax;
double windavg;
uint64_t windPulses = 0;
DHTesp dht;
uint64_t iotTempUpdateTimestamp = 0;
uint64_t iotWindUpdateTimestamp = 0;
uint64_t windUpdateTimestamp = 0;
bool risparmio = false;

//-----------------------CONNESSIONE SERVIZIO MQTT--------------------------------------
void MQTT_connect() {
  int8_t ret;

  // Stop if already connected.
  if (mqtt.connected()) {
    return;
  } else {
    //Serial.println("MQTT Connessione fallita");
  }

  //Serial.print("Connecting to MQTT... ");

  uint8_t retries = 3;
  while ((ret = mqtt.connect()) != 0) { // connect will return 0 for connected
    //Serial.println(mqtt.connectErrorString(ret));
    //Serial.println("Retrying MQTT connection in 5 seconds...");
    mqtt.disconnect();
    delay(5000);  // wait 5 seconds
    retries--;
    if (retries == 0) {
      // basically die and wait for WDT to reset me
      //while (1);
      return;
    }
  }
  Serial.println("MQTT Connected!");
}

//----------------------------LETTTURA TEMPERATURA--------------------------------------
void aggiornaTemperatura() {
  //Lettura temperatura da DHT
  //delay(dht.getMinimumSamplingPeriod());

  humidity = dht.getHumidity();
  temperature = dht.getTemperature();
  temperature = temperature - 2.0;
  temperaturefh = dht.toFahrenheit(temperature);

  Serial.print(dht.getStatusString());
  Serial.print("\t");
  Serial.print(humidity, 1);
  Serial.print("\t\t");
  Serial.print(temperature, 1);
  Serial.print("\t\t");
  Serial.print(dht.toFahrenheit(temperature), 1);
  Serial.print("\t\t");
  Serial.print(dht.computeHeatIndex(temperature, humidity, false), 1);
  Serial.print("\t\t");
  Serial.println(dht.computeHeatIndex(dht.toFahrenheit(temperature), humidity, true), 1);
}

ICACHE_RAM_ATTR static unsigned long last_interrupt_time = 0;

ICACHE_RAM_ATTR void detectsMovement() {
  unsigned long interrupt_time = millis();

  if (interrupt_time - last_interrupt_time > 20)
  {
    ++windPulses;
  }
  last_interrupt_time = interrupt_time;
}

void setup() {
  dht.setup(pinTemp, DHTesp::DHT22); // Connect DHT sensor to GPIO

  WiFi.mode(WIFI_STA);
  WiFi.begin(MySSID, MyWifiPassword);

  // Waiting for Wifi connect
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }

  delay(2000);

  ArduinoOTA.setHostname("ESP-Meteo");
  ArduinoOTA.begin();

  pinMode(pinVento, INPUT_PULLUP);
  attachInterrupt(digitalPinToInterrupt(pinVento), detectsMovement, FALLING);
}

void loop() {
  uint64_t now = millis();

  ArduinoOTA.handle();

  if ((now - iotTempUpdateTimestamp) > IOTTEMP_INTERVAL) {
    // Riaccensione WiFi
    risparmio = true;
    WiFi.forceSleepWake();
    delay(1);
    // Bring up the WiFi connection
    WiFi.mode(WIFI_STA);
    WiFi.begin(MySSID, MyWifiPassword);

    while (WiFi.status() != WL_CONNECTED) {
      delay(100);
    }

    aggiornaTemperatura();
    MQTT_connect();
    iotTempUpdateTimestamp = now;
    FeedTemperatura.publish(temperature);
    FeedUmidita.publish(humidity);
    mqtt.disconnect();
    delay(1000);
    MQTT_connect();
    FeedVento.publish(windavg);
    FeedRaffica.publish(windmax);
    windmax = 0.0;
    //windavg = 0.0;
    mqtt.disconnect();
  } else {
    if (risparmio) {
      WiFi.mode(WIFI_OFF);
      WiFi.forceSleepBegin();
      risparmio = false;
    }
  }
  if ((now - windUpdateTimestamp) > WIND_INTERVAL) {
    windUpdateTimestamp = now;
    wind = (2 * ragAn * PI) * ((windPulses / 1.0) / (WIND_INTERVAL / 1000.0)); //metri al secondo
    wind = wind * 3.6; //chilometri orari
    if (wind > windmax ) {
      windmax = wind;
    }
    windavg = (windavg * filtro) + (wind * (1.0 - filtro));
    windPulses = 0;
  }
}
